/*******************************************************************************
 * Juego de Gato distribuido
 * Implementación de servicios web.
 * Copyright (C) 2013-2014 por Ariel Ortiz
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 ******************************************************************************/

'use strict';

//------------------------------------------------------------------------------
var express    = require('express');
var router     = express.Router();
var async      = require('async');
var mongoose   = require('mongoose');
var constantes = require('../models/constantes.js');
var Juego      = require('../models/juego.js');
var Jugador    = require('../models/jugador.js');

module.exports = router;

//------------------------------------------------------------------------------

var ObjectId   = mongoose.Schema.Types.ObjectId;
var ABORTAR    = true;

//------------------------------------------------------------------------------
router.get('/', function(req, res) {
  res.redirect('/gato/');
});

//------------------------------------------------------------------------------
router.get('/gato/', function (req, res) {
  res.render('index.ejs');
});

//------------------------------------------------------------------------------
router.post('/gato/crear_juego/', function (req, res) {

  var resultado = { creado: false, codigo: 'invalido' };
  var nombre = req.body.nombre;
  var juego;
  var jugador;

  if (nombre) {
    async.waterfall([
      //------------------------------------------------------------------------
      function (callback) {
        Juego.find({ nombre: nombre, iniciado: false }, callback);
      },
      //------------------------------------------------------------------------
      function (juegos, callback) {
        if (juegos.length === 0) {
          juego = new Juego({nombre: nombre});
          juego.save(callback);
        } else {
          resultado.codigo = 'duplicado';
          callback(ABORTAR);
        }
      },
      //------------------------------------------------------------------------
      function (_juego, _n, callback) {
        jugador = new Jugador(
          { juego: juego._id,
            simbolo: constantes.SIMBOLO[0]
          }
        );
        jugador.save(callback);
      },
      //------------------------------------------------------------------------
      function (_jugador, _n, callback) {
        req.session.id_jugador = jugador._id;
        resultado.creado = true;
        resultado.codigo = 'bien';
        resultado.simbolo = jugador.simbolo;
        callback(null);
      }
    ],
    //--------------------------------------------------------------------------
    function (err) {
      if (err && err !== ABORTAR) {
        console.log(err);
      }
      res.json(resultado);
    });
  }
});

//------------------------------------------------------------------------------
router.get('/gato/estado/', function (req, res) {

  var resultado = { estado: 'error'};

  obtenerJuegoJugador(req, function (err, juego, jugador) {

    //--------------------------------------------------------------------------
    function eliminarJuegoJugadores () {
      async.waterfall([
        //----------------------------------------------------------------------
        function (callback) {
          delete req.session.id_jugador;
          jugador.remove(callback);
        },
        //----------------------------------------------------------------------
        function (jugadorBorrado, callback) {
          Jugador.find({ juego: juego._id }, callback);
        },
        //----------------------------------------------------------------------
        function (jugadores, callback) {
          if (jugadores.length === 0) {
            juego.remove(callback);
          } else {
            callback(null);
          }
        }
      ],
      //------------------------------------------------------------------------
      function (err) {
        if (err) {
          console.log(err);
        }
        res.json(resultado);
      });
    };

    //--------------------------------------------------------------------------
    function ganado(s, t) {
      return (s === t[0][0] && s === t[0][1] && s === t[0][2]) ||
             (s === t[1][0] && s === t[1][1] && s === t[1][2]) ||
             (s === t[2][0] && s === t[2][1] && s === t[2][2]) ||
             (s === t[0][0] && s === t[1][0] && s === t[2][0]) ||
             (s === t[0][1] && s === t[1][1] && s === t[2][1]) ||
             (s === t[0][2] && s === t[1][2] && s === t[2][2]) ||
             (s === t[0][0] && s === t[1][1] && s === t[2][2]) ||
             (s === t[0][2] && s === t[1][1] && s === t[2][0]);
    }

    //--------------------------------------------------------------------------
    function lleno(t) {
      for (var i = 0; i < 3; i++) {
        for (var j = 0; j < 3; j++) {
          if (t[i][j] === ' ') return false;
        }
      }
      return true;
    }
    //--------------------------------------------------------------------------

    if (err) {
      console.log(err);
      res.json(resultado);
    } else {

      var tablero = juego.getTablero();

      resultado.tablero = tablero;

      if (!juego.iniciado) {
        resultado.estado = 'espera';
        res.json(resultado);

      } else if (ganado(jugador.simbolo, tablero)) {
        resultado.estado = 'ganaste';
        eliminarJuegoJugadores();

      } else if (ganado(contrincante(jugador.simbolo), tablero)) {
        resultado.estado = 'perdiste';
        eliminarJuegoJugadores();

      } else if (lleno(tablero)) {
        resultado.estado = 'empate';
        eliminarJuegoJugadores();

      } else if (juego.turno === jugador.simbolo) {
        resultado.estado = 'tu_turno';
        res.json(resultado);

      } else {
        resultado.estado = 'espera';
        res.json(resultado);
      }
    }
  });
});

//------------------------------------------------------------------------------
router.get('/gato/juegos_existentes/', function (req, res) {
  Juego
    .find({ iniciado: false })
    .sort('nombre')
    .exec(function (err, juegos) {
      if (err) {
        console.log(err);
      }
      res.json(juegos.map(function (x) {
        return { id: x._id, nombre: x.nombre };
      }));
    });
});

//------------------------------------------------------------------------------
router.put('/gato/tirar/', function (req, res) {

  var resultado = { efectuado: false };

  obtenerJuegoJugador(req, function (err, juego, jugador) {

    //--------------------------------------------------------------------------
    function convertirEntero(s) {
      var r = /^(0*)(\d+)$/.exec(s);
      return r ? parseInt(r[2]) : -1;
    }

    //--------------------------------------------------------------------------
    function guardarCambios(tablero, ren, col) {
      tablero[ren][col] = jugador.simbolo;
      juego.turno = contrincante(juego.turno);
      juego.setTablero(tablero);
      juego.save(function (err) {
        if (err) {
          console.log(err);
        }
        resultado.efectuado = true;
        resultado.tablero = tablero;
        res.json(resultado);
      });
    }

    //--------------------------------------------------------------------------
    function tiroValido(tablero, ren, col) {
      return (0 <= ren && ren <= 2) &&
             (0 <= col && col <= 2) &&
             tablero[ren][col] === ' ';
    }
    //--------------------------------------------------------------------------

    if (err) {
      console.log(err);
      res.json(resultado);
    } else {

      var ren = convertirEntero(req.body.ren);
      var col = convertirEntero(req.body.col);

      if (juego.turno === jugador.simbolo) {

        var tablero = juego.getTablero();

        if (tiroValido(tablero, ren, col)) {
          guardarCambios(tablero, ren, col);
        } else {
          res.json(resultado);
        }
      } else {
        res.json(resultado);
      }
    }
  });
});

//------------------------------------------------------------------------------
router.put('/gato/unir_juego/', function (req, res) {

  var resultado = { unido: false, codigo: 'id_malo' };
  var idJuego = req.body.id_juego;
  var juego;
  var jugador;

  if (idJuego) {
    async.waterfall([
      //------------------------------------------------------------------------
      function (callback) {
        Juego.findOne({_id: idJuego}, callback);
      },
      //------------------------------------------------------------------------
      function (_juego, callback) {
        juego = _juego;
        if (juego.iniciado) {
          callback(ABORTAR);
        } else {
          juego.iniciado = true;
          juego.save(callback);
        }
      },
      //------------------------------------------------------------------------
      function (_juego, _n, callback) {
        jugador = new Jugador(
          { juego: juego._id,
            simbolo: constantes.SIMBOLO[1]
          });
        jugador.save(callback);
      },
      //------------------------------------------------------------------------
      function (_jugador, _n, callback) {
        req.session.id_jugador = jugador._id;
        resultado.unido = true;
        resultado.codigo = 'bien';
        resultado.simbolo = jugador.simbolo;
        callback(null);
      }
    ],
    //--------------------------------------------------------------------------
    function (err) {
      if (err && err !== ABORTAR) {
        console.log(err);
      }
      res.json(resultado);
    });
  } else {
    res.json(resultado);
  }
});

//------------------------------------------------------------------------------
function contrincante(s) {
  return constantes.SIMBOLO[(s === constantes.SIMBOLO[1]) ? 0: 1];
}

//------------------------------------------------------------------------------
function obtenerJuegoJugador(req, callback) {

  var idJugador = req.session.id_jugador;
  var juego;
  var jugador;

  if (idJugador) {
    async.waterfall([
      //------------------------------------------------------------------------
      function (callback) {
        Jugador.findOne({ _id: idJugador }, callback);
      },
      //------------------------------------------------------------------------
      function (_jugador, callback) {
        jugador = _jugador;
        Juego.findOne({ _id: jugador.juego }, callback);
      },
      //------------------------------------------------------------------------
      function (_juego, callback) {
        juego = _juego;
        callback(null);
      }
    ],
    //--------------------------------------------------------------------------
    function (err) {
      if (err) {
        console.log(err);
      }
      callback(null, juego, jugador);
    });
  } else {
    callback(Error('La sesión no contiene el ID del jugador'));
  }
}
